"""Test shading rules."""

import collections
import copy
import datetime
import pathlib
import sys
import time
import unittest
import unittest.mock

import HABApp.rule.rule
import pydantic_core

import habapp_rules.actors.config.shading
import habapp_rules.actors.light
import habapp_rules.actors.shading
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer
from habapp_rules.system import PresenceState


class TestShadingBase(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing _ShadingBase."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_Shading_min", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_min", "OFF")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Shading_max", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_max", "OFF")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Shading_min_state", "")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "CustomState", "")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_WindAlarm", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_SunProtection", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", habapp_rules.system.SleepState.AWAKE.value)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Night", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door", "CLOSED")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Hand_Manual_active", "OFF")

        config_min = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading_min", manual="Unittest_Manual_min", state="H_Unittest_Shading_min_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
        )

        config_max = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading_max",
                manual="Unittest_Manual_max",
                wind_alarm="Unittest_WindAlarm",
                sun_protection="Unittest_SunProtection",
                sleeping_state="Unittest_Sleep_state",
                night="Unittest_Night",
                door="Unittest_Door",
                summer="Unittest_Summer",
                hand_manual_is_active_feedback="Unittest_Hand_Manual_active",
                state="CustomState",
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )

        self.shading_min = habapp_rules.actors.shading._ShadingBase(config_min)
        self.shading_max = habapp_rules.actors.shading._ShadingBase(config_max)

    def test__init__(self) -> None:
        """Test __init__."""
        expected_states = [
            {"name": "WindAlarm"},
            {"name": "Manual"},
            {"name": "Hand", "on_timeout": "_auto_hand_timeout", "timeout": 72000},
            {
                "name": "Auto",
                "initial": "Init",
                "children": [
                    {"name": "Init"},
                    {"name": "Open"},
                    {"name": "DoorOpen", "initial": "Open", "children": [{"name": "Open"}, {"name": "PostOpen", "on_timeout": "_timeout_post_door_open", "timeout": 300}]},
                    {"name": "NightClose"},
                    {"name": "SleepingClose"},
                    {"name": "SunProtection"},
                ],
            },
        ]
        self.assertEqual(expected_states, self.shading_max.states)

        expected_trans = [
            {"trigger": "_wind_alarm_on", "source": ["Auto", "Hand", "Manual"], "dest": "WindAlarm"},
            {"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Manual", "conditions": "_manual_active"},
            {"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Auto", "unless": "_manual_active"},
            # manual
            {"trigger": "_manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
            {"trigger": "_manual_off", "source": "Manual", "dest": "Auto"},
            # hand
            {"trigger": "_hand_command", "source": ["Auto"], "dest": "Hand"},
            {"trigger": "_auto_hand_timeout", "source": "Hand", "dest": "Auto"},
            # sun
            {"trigger": "_sun_on", "source": "Auto_Open", "dest": "Auto_SunProtection"},
            {"trigger": "_sun_off", "source": "Auto_SunProtection", "dest": "Auto_Open"},
            # sleep
            {"trigger": "_sleep_started", "source": ["Auto_Open", "Auto_NightClose", "Auto_SunProtection", "Auto_DoorOpen"], "dest": "Auto_SleepingClose"},
            {"trigger": "_sleep_started", "source": "Hand", "dest": "Auto"},
            {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
            {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_NightClose", "conditions": ["_night_active_and_configured"]},
            {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_Open", "unless": ["_night_active_and_configured", "_sun_protection_active_and_configured"]},
            # door
            {"trigger": "_door_open", "source": ["Auto_NightClose", "Auto_SunProtection", "Auto_SleepingClose", "Auto_Open"], "dest": "Auto_DoorOpen"},
            {"trigger": "_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_DoorOpen_Open"},
            {"trigger": "_door_closed", "source": "Auto_DoorOpen_Open", "dest": "Auto_DoorOpen_PostOpen"},
            {"trigger": "_timeout_post_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_Init"},
            # night close
            {"trigger": "_night_started", "source": ["Auto_Open", "Auto_SunProtection"], "dest": "Auto_NightClose", "conditions": "_night_active_and_configured"},
            {"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
            {"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_Open", "unless": ["_sun_protection_active_and_configured"]},
        ]

        self.assertEqual(expected_trans, self.shading_max.trans)

        self.assertEqual(300, self.shading_max.state_machine.states["Auto"].states["DoorOpen"].states["PostOpen"].timeout)
        self.assertEqual(86400, self.shading_max.state_machine.states["Manual"].timeout)

    def test_init_exceptions(self) -> None:
        """Test exceptions of __init__."""
        TestCase = collections.namedtuple("TestCase", "item_type, raises_exc")

        test_cases = [
            TestCase(HABApp.openhab.items.RollershutterItem, False),
            TestCase(HABApp.openhab.items.DimmerItem, False),
            TestCase(HABApp.openhab.items.SwitchItem, True),
            TestCase(HABApp.openhab.items.ContactItem, True),
            TestCase(HABApp.openhab.items.NumberItem, True),
        ]

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Temp_state", "")

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.add_mock_item(test_case.item_type, "Unittest_Temp", None)

                if test_case.raises_exc:
                    with self.assertRaises(pydantic_core.ValidationError):
                        habapp_rules.actors.config.shading.ShadingConfig(
                            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Temp", manual="Unittest_Manual_min", state="H_Unittest_Shading_min_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
                        )
                else:
                    config = habapp_rules.actors.config.shading.ShadingConfig(
                        items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Temp", manual="Unittest_Manual_min", state="H_Unittest_Shading_min_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
                    )
                    habapp_rules.actors.shading._ShadingBase(config)
                tests.helper.oh_item.remove_mocked_item_by_name("Unittest_Temp")

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Shading"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        jal_graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.shading_min.states, transitions=self.shading_min.trans, initial=self.shading_min.state, show_conditions=False)

        jal_graph.get_graph().draw(picture_dir / "Shading.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self.shading_min.states) if "init" not in state.lower()]:
            jal_graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.shading_min.states, transitions=self.shading_min.trans, initial=state_name, show_conditions=True)
            jal_graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"Shading_{state_name}.png", format="png", prog="dot")

    @staticmethod
    def get_initial_state_test_cases() -> list[collections.namedtuple]:
        """Get test cases for initial state tests.

        Returns:
            tests cases
        """
        TestCase = collections.namedtuple("TestCase", "wind_alarm, manual, sleeping_state, door, night, sun_protection, expected_state")

        return [
            # wind_alarm = OFF | manual = OFF
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="Auto_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="ON", expected_state="Auto_SunProtection"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="OFF", expected_state="Auto_NightClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="ON", expected_state="Auto_NightClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="OFF", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="ON", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="OFF", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="ON", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="Auto_SleepingClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="ON", expected_state="Auto_SleepingClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="OFF", expected_state="Auto_SleepingClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="ON", expected_state="Auto_SleepingClose"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="OFF", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="ON", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="OFF", expected_state="Auto_DoorOpen_Open"),
            TestCase(wind_alarm="OFF", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="ON", expected_state="Auto_DoorOpen_Open"),
            # wind_alarm = OFF | manual = ON
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="ON", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="OFF", expected_state="Manual"),
            TestCase(wind_alarm="OFF", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="ON", expected_state="Manual"),
            # wind_alarm = ON | manual = OFF
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="OFF", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            # wind_alarm = ON | manual = ON
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="CLOSED", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.AWAKE, door="OPEN", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="CLOSED", night="ON", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="OFF", sun_protection="ON", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="OFF", expected_state="WindAlarm"),
            TestCase(wind_alarm="ON", manual="ON", sleeping_state=habapp_rules.system.SleepState.SLEEPING, door="OPEN", night="ON", sun_protection="ON", expected_state="WindAlarm"),
        ]

    def test_get_initial_state(self) -> None:
        """Test _get_initial_state."""
        for test_case in self.get_initial_state_test_cases():
            tests.helper.oh_item.set_state("Unittest_WindAlarm", test_case.wind_alarm)
            tests.helper.oh_item.set_state("Unittest_Manual_max", test_case.manual)
            tests.helper.oh_item.set_state("Unittest_Sleep_state", test_case.sleeping_state.value)
            tests.helper.oh_item.set_state("Unittest_Door", test_case.door)
            tests.helper.oh_item.set_state("Unittest_Night", test_case.night)
            tests.helper.oh_item.set_state("Unittest_SunProtection", test_case.sun_protection)

            self.assertEqual(test_case.expected_state, self.shading_max._get_initial_state())

    @staticmethod
    def get_target_positions_test_cases() -> list[collections.namedtuple]:
        """Get test cases for target position.

        Returns:
            tests cases
        """
        TestCase = collections.namedtuple("TestCase", "state, target_pos")

        return [
            TestCase("Hand", None),
            TestCase("Manual", None),
            TestCase("WindAlarm", habapp_rules.actors.config.shading.ShadingPosition(0, 0)),
            TestCase("Auto_Open", habapp_rules.actors.config.shading.ShadingPosition(0, 0)),
            TestCase("Auto_SunProtection", habapp_rules.actors.config.shading.ShadingPosition(100, None)),
            TestCase("Auto_SleepingClose", habapp_rules.actors.config.shading.ShadingPosition(100, 100)),
            TestCase("Auto_NightClose", habapp_rules.actors.config.shading.ShadingPosition(100, 100)),
            TestCase("Auto_DoorOpen_Open", habapp_rules.actors.config.shading.ShadingPosition(0, 0)),
            TestCase("Auto_DoorOpen_PostOpen", None),
        ]

    def test_get_target_position(self) -> None:
        """Test _get_target_position."""
        for test_case in self.get_target_positions_test_cases():
            with self.subTest(test_case=test_case):
                self.shading_max._set_state(test_case.state)
                self.assertEqual(test_case.target_pos, self.shading_max._get_target_position())

        tests.helper.oh_item.send_command("Unittest_Summer", "ON", "OFF")
        self.shading_max._set_state("Auto_NightClose")
        self.assertEqual(None, self.shading_max._get_target_position())

    def test_get_target_position_sleeping(self) -> None:
        """Test get_target_position if sleeping is active."""
        config_night = habapp_rules.actors.config.shading.ShadingPosition(20, 30)
        config_day = habapp_rules.actors.config.shading.ShadingPosition(40, 50)

        # item night is not None
        with unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_night", config_night), unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_day", config_day):
            self.shading_max.state = "Auto_SleepingClose"
            tests.helper.oh_item.set_state("Unittest_Night", "ON")
            self.assertEqual(config_night, self.shading_max._get_target_position())

            tests.helper.oh_item.set_state("Unittest_Night", "OFF")
            self.assertEqual(config_day, self.shading_max._get_target_position())

        # item night is None
        with (
            unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_night", config_night),
            unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_day", config_day),
            unittest.mock.patch.object(self.shading_max._config.items, "night", None),
        ):
            self.shading_max.state = "Auto_SleepingClose"
            tests.helper.oh_item.set_state("Unittest_Night", "ON")
            self.assertEqual(config_night, self.shading_max._get_target_position())

            tests.helper.oh_item.set_state("Unittest_Night", "OFF")
            self.assertEqual(config_night, self.shading_max._get_target_position())

    def test_cb_sleep_state(self) -> None:
        """Test _cb_sleep_state."""
        TestCase = collections.namedtuple("TestCase", "sleep_state, started_triggered, stopped_triggered")

        test_cases = [
            TestCase(habapp_rules.system.SleepState.AWAKE, False, False),
            TestCase(habapp_rules.system.SleepState.PRE_SLEEPING, True, False),
            TestCase(habapp_rules.system.SleepState.SLEEPING, False, False),
            TestCase(habapp_rules.system.SleepState.POST_SLEEPING, False, True),
            TestCase(habapp_rules.system.SleepState.LOCKED, False, False),
        ]

        with unittest.mock.patch.object(self.shading_max, "_sleep_started") as started_mock, unittest.mock.patch.object(self.shading_max, "_sleep_stopped") as stopped_mock:
            for test_case in test_cases:
                started_mock.reset_mock()
                stopped_mock.reset_mock()

                tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", test_case.sleep_state.value)

                if test_case.started_triggered:
                    started_mock.assert_called_once()
                else:
                    started_mock.assert_not_called()

                if test_case.stopped_triggered:
                    stopped_mock.assert_called_once()
                else:
                    stopped_mock.assert_not_called()

    def test_cb_night(self) -> None:
        """Test _cb_night."""
        config_night = habapp_rules.actors.config.shading.ShadingPosition(20, 30)
        config_day = habapp_rules.actors.config.shading.ShadingPosition(40, 50)

        self.shading_max.state = "Auto_SleepingClose"
        with unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_night", config_night), unittest.mock.patch.object(self.shading_max._config.parameter, "pos_sleeping_day", config_day):
            with unittest.mock.patch.object(self.shading_max, "_apply_target_position") as apply_pos_mock:
                tests.helper.oh_item.item_state_change_event("Unittest_Night", "OFF")
            apply_pos_mock.assert_called_once_with(config_day)

            with unittest.mock.patch.object(self.shading_max, "_apply_target_position") as apply_pos_mock:
                tests.helper.oh_item.item_state_change_event("Unittest_Night", "ON")
            apply_pos_mock.assert_called_once_with(config_night)

            self.shading_max.state = "Manual"
            with unittest.mock.patch.object(self.shading_max, "_apply_target_position") as apply_pos_mock:
                tests.helper.oh_item.item_state_change_event("Unittest_Night", "OFF")
            apply_pos_mock.assert_not_called()

    def test_cb_hand(self) -> None:
        """Test _cb_hand."""
        self.shading_min.to_Auto()
        self.shading_max.to_Auto()

        # last rule command was less than a second ago
        self.shading_min._set_shading_state_timestamp = time.time() - 0.1
        self.shading_max._set_shading_state_timestamp = time.time() - 0.1

        self.shading_min._cb_hand(unittest.mock.MagicMock())
        self.shading_max._cb_hand(unittest.mock.MagicMock())

        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_Open", self.shading_max.state)

        # last rule command was longer ago
        self.shading_min._set_shading_state_timestamp = time.time() - 1.6
        self.shading_max._set_shading_state_timestamp = time.time() - 1.6

        self.shading_min._cb_hand(unittest.mock.MagicMock())
        self.shading_max._cb_hand(unittest.mock.MagicMock())

        self.assertEqual("Hand", self.shading_min.state)
        self.assertEqual("Hand", self.shading_max.state)

    def test_night_active_and_configured(self) -> None:
        """Test _night_active_and_configured."""
        TestCase = collections.namedtuple("TestCase", "night, summer, config_summer, config_winter, expected_result")

        shading_pos = habapp_rules.actors.config.shading.ShadingPosition(42, 0)

        test_cases = [
            # night off
            TestCase("OFF", None, None, None, False),
            TestCase("OFF", None, None, shading_pos, False),
            TestCase("OFF", None, shading_pos, None, False),
            TestCase("OFF", None, shading_pos, shading_pos, False),
            TestCase("OFF", "OFF", None, None, False),
            TestCase("OFF", "OFF", None, shading_pos, False),
            TestCase("OFF", "OFF", shading_pos, None, False),
            TestCase("OFF", "OFF", shading_pos, shading_pos, False),
            TestCase("OFF", "ON", None, None, False),
            TestCase("OFF", "ON", None, shading_pos, False),
            TestCase("OFF", "ON", shading_pos, None, False),
            TestCase("OFF", "ON", shading_pos, shading_pos, False),
            # night on
            TestCase("ON", None, None, None, False),
            TestCase("ON", None, None, shading_pos, True),
            TestCase("ON", None, shading_pos, None, False),
            TestCase("ON", None, shading_pos, shading_pos, True),
            TestCase("ON", "OFF", None, None, False),
            TestCase("ON", "OFF", None, shading_pos, True),
            TestCase("ON", "OFF", shading_pos, None, False),
            TestCase("ON", "OFF", shading_pos, shading_pos, True),
            TestCase("ON", "ON", None, None, False),
            TestCase("ON", "ON", None, shading_pos, False),
            TestCase("ON", "ON", shading_pos, None, True),
            TestCase("ON", "ON", shading_pos, shading_pos, True),
        ]

        for test_case in test_cases:
            with unittest.mock.patch.object(self.shading_max._config.parameter, "pos_night_close_summer", test_case.config_summer), unittest.mock.patch.object(self.shading_max._config.parameter, "pos_night_close_winter", test_case.config_winter):
                tests.helper.oh_item.set_state("Unittest_Summer", test_case.summer)
                tests.helper.oh_item.set_state("Unittest_Night", test_case.night)

                self.assertEqual(test_case.expected_result, self.shading_max._night_active_and_configured())

    def test_manual_transitions(self) -> None:
        """Test transitions of state manual."""
        for initial_state in ("Hand", "Auto"):
            self.shading_min.state_machine.set_state(initial_state)
            self.shading_max.state_machine.set_state(initial_state)
            tests.helper.oh_item.item_state_change_event("Unittest_Manual_min", "ON", "OFF")
            tests.helper.oh_item.item_state_change_event("Unittest_Manual_max", "ON", "OFF")
            self.assertEqual("Manual", self.shading_min.state)
            self.assertEqual("Manual", self.shading_max.state)

        # to WindAlarm | without wind_alarm_item
        self.shading_min.to_Manual()
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "ON", "OFF")
        self.assertEqual("Manual", self.shading_min.state)

        # to WindAlarm | with wind_alarm_item (and back to manual -> states must be the same as before)
        self.shading_max.to_Manual()
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "ON", "OFF")
        self.assertEqual("WindAlarm", self.shading_max.state)

        # to Auto (manual off)
        self.shading_min.to_Manual()
        self.shading_max.to_Manual()
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_Manual_min", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_Manual_max", "OFF", "ON")
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_Open", self.shading_max.state)

    def test_hand_transitions(self) -> None:
        """Test transitions of state hand."""
        # from auto to hand
        self.shading_min.to_Auto()
        self.shading_max.to_Auto()
        self.shading_min._hand_command()
        self.shading_max._hand_command()
        self.assertEqual("Hand", self.shading_min.state)
        self.assertEqual("Hand", self.shading_max.state)

        # from hand to auto
        self.shading_min.to_Hand()
        self.shading_max.to_Hand()
        self.shading_min._auto_hand_timeout()
        self.shading_max._auto_hand_timeout()
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_Open", self.shading_max.state)

        # from hand to manual
        self.shading_min.to_Hand()
        self.shading_max.to_Hand()
        tests.helper.oh_item.send_command("Unittest_Manual_min", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Manual_max", "ON", "OFF")
        self.assertEqual("Manual", self.shading_min.state)
        self.assertEqual("Manual", self.shading_max.state)

        # from hand to wind_alarm
        self.shading_min.to_Hand()
        self.shading_max.to_Hand()
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "ON", "OFF")
        self.assertEqual("Hand", self.shading_min.state)
        self.assertEqual("WindAlarm", self.shading_max.state)

    def test_wind_alarm_transitions(self) -> None:
        """Test transitions of state WindAlarm."""
        # from wind_alarm to manual
        self.shading_max.to_WindAlarm()
        tests.helper.oh_item.send_command("Unittest_Manual_max", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "OFF", "ON")
        self.assertEqual("Manual", self.shading_max.state)

        # from wind_alarm to Auto
        self.shading_max.to_WindAlarm()
        tests.helper.oh_item.send_command("Unittest_Manual_max", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_WindAlarm", "OFF", "ON")
        self.assertEqual("Auto_Open", self.shading_max.state)

    def test_auto_open_transitions(self) -> None:
        """Test transitions of state Auto_Open."""
        # to door_open
        self.shading_min.to_Auto_Open()
        self.shading_max.to_Auto_Open()
        tests.helper.oh_item.send_command("Unittest_Door", "OPEN", "CLOSED")
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_DoorOpen_Open", self.shading_max.state)

        # to night_close | configured
        self.shading_min.to_Auto_Open()
        self.shading_max.to_Auto_Open()
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_NightClose", self.shading_max.state)

        # to night_close | NOT configured
        self.shading_min.to_Auto_Open()
        self.shading_max.to_Auto_Open()
        tests.helper.oh_item.send_command("Unittest_Summer", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to sleeping_close
        self.shading_min.to_Auto_Open()
        self.shading_max.to_Auto_Open()
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_SleepingClose", self.shading_max.state)

        # to sun_protection
        self.shading_min.to_Auto_Open()
        self.shading_max.to_Auto_Open()
        tests.helper.oh_item.send_command("Unittest_SunProtection", "ON", "OFF")
        self.assertEqual("Auto_Open", self.shading_min.state)
        self.assertEqual("Auto_SunProtection", self.shading_max.state)

    def test_auto_night_close_transitions(self) -> None:
        """Test transitions of state Auto_NightClose."""
        # to open
        self.shading_max.to_Auto_NightClose()
        tests.helper.oh_item.send_command("Unittest_SunProtection", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_Night", "OFF", "ON")
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to door_open
        self.shading_max.to_Auto_NightClose()
        tests.helper.oh_item.send_command("Unittest_Door", "OPEN", "CLOSED")
        self.assertEqual("Auto_DoorOpen_Open", self.shading_max.state)

        # to sleeping_close
        self.shading_max.to_Auto_NightClose()
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_SleepingClose", self.shading_max.state)

        # to sun_protection
        self.shading_max.to_Auto_NightClose()
        tests.helper.oh_item.send_command("Unittest_SunProtection", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Night", "OFF", "ON")
        self.assertEqual("Auto_SunProtection", self.shading_max.state)

    def test_auto_sleeping_close_transitions(self) -> None:
        """Test transitions of state Auto_SleepingClose."""
        # to open
        self.shading_max.to_Auto_SleepingClose()
        tests.helper.oh_item.send_command("Unittest_Night", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_SunProtection", "OFF", "OFF")
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to night_close | configured
        self.shading_max.to_Auto_SleepingClose()
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("Auto_NightClose", self.shading_max.state)

        # to night_close | NOT configured
        self.shading_max.to_Auto_SleepingClose()
        tests.helper.oh_item.send_command("Unittest_Summer", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to door_open
        self.shading_max.to_Auto_SleepingClose()
        tests.helper.oh_item.send_command("Unittest_Door", "OPEN", "CLOSED")
        self.assertEqual("Auto_DoorOpen_Open", self.shading_max.state)

        # to sun_protection
        self.shading_max.to_Auto_SleepingClose()
        tests.helper.oh_item.send_command("Unittest_SunProtection", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("Auto_SunProtection", self.shading_max.state)

    def test_auto_sun_protection_transitions(self) -> None:
        """Test transitions of state Auto_SunProtection."""
        # to open
        self.shading_max.to_Auto_SunProtection()
        tests.helper.oh_item.send_command("Unittest_SunProtection", "OFF", "ON")
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to night_close | configured
        self.shading_max.to_Auto_SunProtection()
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        self.assertEqual("Auto_NightClose", self.shading_max.state)

        # to night_close | NOT configured
        self.shading_max.to_Auto_SunProtection()
        tests.helper.oh_item.send_command("Unittest_Summer", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Night", "ON", "OFF")
        self.assertEqual("Auto_SunProtection", self.shading_max.state)

        # to sleeping_protection
        self.shading_max.to_Auto_SunProtection()
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_SleepingClose", self.shading_max.state)

        # to door_open
        self.shading_max.to_Auto_SunProtection()
        tests.helper.oh_item.send_command("Unittest_Door", "OPEN", "CLOSED")
        self.assertEqual("Auto_DoorOpen_Open", self.shading_max.state)

    def test_auto_door_open_transitions(self) -> None:
        """Test transitions of state Auto_DoorOpen."""
        # door closed -> PostOpen
        self.shading_max.to_Auto_DoorOpen_Open()
        tests.helper.oh_item.send_command("Unittest_Door", "CLOSED", "OPEN")
        self.assertEqual("Auto_DoorOpen_PostOpen", self.shading_max.state)

        # door opened again -> Open
        tests.helper.oh_item.send_command("Unittest_Door", "OPEN", "CLOSED")
        self.assertEqual("Auto_DoorOpen_Open", self.shading_max.state)

        # door closed + timeout -> AutoInit
        tests.helper.oh_item.send_command("Unittest_Door", "CLOSED", "OPEN")
        self.shading_max._timeout_post_door_open()
        self.assertEqual("Auto_Open", self.shading_max.state)

        # to sleeping from open
        self.shading_max.to_Auto_DoorOpen_Open()
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_SleepingClose", self.shading_max.state)

        # to sleeping from post open
        self.shading_max.to_Auto_DoorOpen_PostOpen()
        tests.helper.oh_item.send_command("Unittest_Sleep_state", habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.AWAKE.value)
        self.assertEqual("Auto_SleepingClose", self.shading_max.state)


class TestShadingShutter(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing Raffstore class."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Shading", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual", "OFF")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Shading_state", "")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_WindAlarm", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_SunProtection", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", habapp_rules.system.SleepState.AWAKE.value)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Night", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door", "CLOSED")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Hand_Manual_active", "OFF")

        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading",
                manual="Unittest_Manual",
                wind_alarm="Unittest_WindAlarm",
                sun_protection="Unittest_SunProtection",
                sleep_state="Unittest_Sleep_state",
                night="Unittest_Night",
                door="Unittest_Door",
                summer="Unittest_Summer",
                hand_manual_active="Unittest_Hand_Manual_active",
                state="H_Unittest_Shading_state",
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )

        self.shutter = habapp_rules.actors.shading.Shutter(config)

    def test_set_shading_state(self) -> None:
        """Test _set_shading_state."""
        TestCase = collections.namedtuple("TestCase", "target_pos, sent_pos")
        self.shutter._set_shading_state_timestamp = None

        test_cases = [
            TestCase(None, None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(None, None), None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(None, 25), None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(50, None), 50),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(80, 90), 80),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(0, 0), 0),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(0, 25), 0),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(50, 0), 50),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(80, 90), 80),
        ]

        with unittest.mock.patch.object(self.shutter, "_get_target_position") as get_pos_mock, unittest.mock.patch.object(self.shutter._state_observer_pos, "send_command") as send_pos_mock:
            for test_case in test_cases:
                get_pos_mock.return_value = test_case.target_pos
                send_pos_mock.reset_mock()

                self.shutter._set_shading_state()

                if test_case.sent_pos is None:
                    send_pos_mock.assert_not_called()
                else:
                    send_pos_mock.assert_called_once_with(test_case.sent_pos)

        self.assertTrue(self.shutter._set_shading_state_timestamp > 1000)


class TestShadingRaffstore(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing Raffstore class."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_Shading", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Slat", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual", "OFF")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Shading_state", "")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_WindAlarm", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_SunProtection", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_SunProtection_Slat", 83)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", habapp_rules.system.SleepState.AWAKE.value)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Night", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door", "CLOSED")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Hand_Manual_active", "OFF")

        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading",
                slat="Unittest_Slat",
                manual="Unittest_Manual",
                wind_alarm="Unittest_WindAlarm",
                sun_protection="Unittest_SunProtection",
                sun_protection_slat="Unittest_SunProtection_Slat",
                sleeping_state="Unittest_Sleep_state",
                night="Unittest_Night",
                door="Unittest_Door",
                summer="Unittest_Summer",
                hand_manual_active="Unittest_Hand_Manual_active",
                state="H_Unittest_Shading_state",
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )

        self.raffstore = habapp_rules.actors.shading.Raffstore(config)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_Shading", None)
        tests.helper.oh_item.set_state("Unittest_Slat", None)
        tests.helper.oh_item.set_state("Unittest_Manual", None)
        tests.helper.oh_item.set_state("Unittest_WindAlarm", None)
        tests.helper.oh_item.set_state("Unittest_SunProtection", None)
        tests.helper.oh_item.set_state("Unittest_SunProtection_Slat", None)
        tests.helper.oh_item.set_state("Unittest_Sleep_state", None)
        tests.helper.oh_item.set_state("Unittest_Night", None)
        tests.helper.oh_item.set_state("Unittest_Door", None)
        tests.helper.oh_item.set_state("Unittest_Summer", None)
        tests.helper.oh_item.set_state("Unittest_Hand_Manual_active", None)

        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading",
                slat="Unittest_Slat",
                manual="Unittest_Manual",
                wind_alarm="Unittest_WindAlarm",
                sun_protection="Unittest_SunProtection",
                sun_protection_slat="Unittest_SunProtection_Slat",
                sleeping_state="Unittest_Sleep_state",
                night="Unittest_Night",
                door="Unittest_Door",
                summer="Unittest_Summer",
                hand_manual_active="Unittest_Hand_Manual_active",
                state="H_Unittest_Shading_state",
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )

        habapp_rules.actors.shading.Raffstore(config)

    def test_init_min(self) -> None:
        """Test init of raffstore with minimal attributes."""
        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", slat="Unittest_Slat", manual="Unittest_Manual", state="H_Unittest_Shading_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
        )

        habapp_rules.actors.shading.Raffstore(config)

    def test_init_missing_items(self) -> None:
        """Test init with missing items."""
        # sun_protection item NOT given | item sun_protection_slat NOT given
        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", slat="Unittest_Slat", manual="Unittest_Manual", state="H_Unittest_Shading_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
        )
        habapp_rules.actors.shading.Raffstore(config)

        # sun_protection item NOT given | item sun_protection_slat given
        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", slat="Unittest_Slat", manual="Unittest_Manual", state="H_Unittest_Shading_state", sun_protection_slat="Unittest_SunProtection_Slat"),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.shading.Raffstore(config)

        # sun_protection item given | item sun_protection_slat NOT given
        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading",
                slat="Unittest_Slat",
                manual="Unittest_Manual",
                state="H_Unittest_Shading_state",
                sun_protection="Unittest_SunProtection",
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.shading.Raffstore(config)

        # sun_protection item given | item sun_protection_slat given
        config = habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(
                shading_position="Unittest_Shading", slat="Unittest_Slat", manual="Unittest_Manual", state="H_Unittest_Shading_state", sun_protection="Unittest_SunProtection", sun_protection_slat="Unittest_SunProtection_Slat"
            ),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )
        habapp_rules.actors.shading.Raffstore(config)

        # slat is missing
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            config = habapp_rules.actors.config.shading.ShadingConfig(
                items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", manual="Unittest_Manual", state="H_Unittest_Shading_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
            )
            habapp_rules.actors.shading.Raffstore(config)

    def test_verify_items(self) -> None:
        """Test __verify_items."""
        # test shading position type
        TestCase = collections.namedtuple("TestCase", "item_type, raises_exc")

        test_cases = [
            TestCase(HABApp.openhab.items.RollershutterItem, False),
            TestCase(HABApp.openhab.items.DimmerItem, True),
            TestCase(HABApp.openhab.items.SwitchItem, True),
            TestCase(HABApp.openhab.items.ContactItem, True),
            TestCase(HABApp.openhab.items.NumberItem, True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.raffstore._config.items.shading_position = test_case.item_type("Name")
                if test_case.raises_exc:
                    with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
                        self.raffstore._Raffstore__verify_items()
                else:
                    self.raffstore._Raffstore__verify_items()

    def test_get_target_position(self) -> None:
        """Test _get_target_position."""
        TestCase = collections.namedtuple("TestCase", "state, target_pos")

        test_cases = TestShadingBase.get_target_positions_test_cases()
        test_cases[4] = TestCase("Auto_SunProtection", habapp_rules.actors.config.shading.ShadingPosition(100, 83))

        for test_case in test_cases:
            self.raffstore._set_state(test_case.state)
            self.assertEqual(test_case.target_pos, self.raffstore._get_target_position())

    def test_set_shading_state(self) -> None:
        """Test _set_shading_state."""
        TestCase = collections.namedtuple("TestCase", "target_pos, sent_pos, sent_slat")

        test_cases = [
            TestCase(None, None, None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(None, None), None, None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(None, 25), None, 25),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(50, None), 50, None),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(80, 90), 80, 90),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(0, 0), 0, 0),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(0, 25), 0, 25),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(50, 0), 50, 0),
            TestCase(habapp_rules.actors.config.shading.ShadingPosition(80, 90), 80, 90),
        ]

        with (
            unittest.mock.patch.object(self.raffstore, "_get_target_position") as get_pos_mock,
            unittest.mock.patch.object(self.raffstore._state_observer_pos, "send_command") as send_pos_mock,
            unittest.mock.patch.object(self.raffstore._state_observer_slat, "send_command") as send_slat_mock,
        ):
            for test_case in test_cases:
                get_pos_mock.return_value = test_case.target_pos
                send_pos_mock.reset_mock()
                send_slat_mock.reset_mock()

                self.raffstore._set_shading_state()

                if test_case.sent_pos is None:
                    send_pos_mock.assert_not_called()
                else:
                    send_pos_mock.assert_called_once_with(test_case.sent_pos)

                if test_case.sent_slat is None:
                    send_slat_mock.assert_not_called()
                else:
                    send_slat_mock.assert_called_once_with(test_case.sent_slat)

    def test_cb_slat(self) -> None:
        """Test _cb_slat."""
        with unittest.mock.patch.object(self.raffstore._state_observer_slat, "send_command") as send_command_mock:
            tests.helper.oh_item.send_command("Unittest_Manual", "ON", "OFF")
            tests.helper.oh_item.send_command("Unittest_SunProtection_Slat", 15, 99)
            send_command_mock.assert_not_called()

            tests.helper.oh_item.send_command("Unittest_Manual", "OFF", "ON")
            tests.helper.oh_item.send_command("Unittest_SunProtection", "ON", "OFF")
            self.assertEqual("Auto_SunProtection", self.raffstore.state)
            tests.helper.oh_item.send_command("Unittest_SunProtection_Slat", 22, 15)
            send_command_mock.assert_has_calls([unittest.mock.call(0), unittest.mock.call(15), unittest.mock.call(22)])

    def test_manual_position_restore(self) -> None:
        """Test if manual position is restored correctly."""
        tests.helper.oh_item.send_command("Unittest_Manual", "ON", "OFF")

        tests.helper.oh_item.send_command("Unittest_Shading", 12)
        tests.helper.oh_item.send_command("Unittest_Slat", 34)

        tests.helper.oh_item.send_command("Unittest_WindAlarm", "ON", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Shading", 0)
        tests.helper.oh_item.assert_value("Unittest_Slat", 0)

        tests.helper.oh_item.send_command("Unittest_WindAlarm", "OFF", "ON")
        tests.helper.oh_item.assert_value("Unittest_Shading", 12)
        tests.helper.oh_item.assert_value("Unittest_Slat", 34)


class TestResetAllManualHand(tests.helper.test_case_base.TestCaseBase):
    """Tests for ResetAllManualHand."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Reset", None)

        config = habapp_rules.actors.config.shading.ResetAllManualHandConfig(items=habapp_rules.actors.config.shading.ResetAllManualHandItems(reset_manual_hand="Unittest_Reset"))

        self.reset_shading_rule = habapp_rules.actors.shading.ResetAllManualHand(config)

    def test__get_shading_objects(self) -> None:
        """Test __get_shading_objects."""
        # shading objects where given
        shading_object_1 = unittest.mock.MagicMock()
        shading_object_2 = unittest.mock.MagicMock()

        config_2 = habapp_rules.actors.config.shading.ResetAllManualHandConfig(
            items=habapp_rules.actors.config.shading.ResetAllManualHandItems(
                reset_manual_hand="Unittest_Reset",
            ),
            parameter=habapp_rules.actors.config.shading.ResetAllManualHandParameter(shading_objects=[shading_object_1, shading_object_2]),
        )

        reset_shading_rule_2 = habapp_rules.actors.shading.ResetAllManualHand(config_2)
        self.assertEqual([shading_object_1, shading_object_2], reset_shading_rule_2._ResetAllManualHand__get_shading_objects())

        # shading objects are not set via __init__
        with unittest.mock.patch.object(self.reset_shading_rule, "get_rule") as get_rule_mock:
            self.reset_shading_rule._ResetAllManualHand__get_shading_objects()

        get_rule_mock.assert_called_once_with(None)

    def test_cb_reset_all(self) -> None:
        """Test _cb_reset_all."""
        TestCase = collections.namedtuple("TestCase", "event, state, manual_commands")

        test_cases = [
            TestCase("OFF", "Auto", []),
            TestCase("OFF", "Hand", []),
            TestCase("OFF", "Manual", []),
            TestCase("ON", "Auto", []),
            TestCase("ON", "Hand", [unittest.mock.call("ON"), unittest.mock.call("OFF")]),
            TestCase("ON", "Manual", [unittest.mock.call("OFF")]),
        ]

        shading_rule_mock = unittest.mock.MagicMock()
        shading_rule_mock._config.items.manual = unittest.mock.MagicMock(spec=HABApp.openhab.items.SwitchItem)

        with unittest.mock.patch.object(self.reset_shading_rule, "_ResetAllManualHand__get_shading_objects", return_value=[shading_rule_mock]):
            for test_case in test_cases:
                shading_rule_mock.state = test_case.state
                shading_rule_mock._config.items.manual.oh_send_command.reset_mock()

                self.reset_shading_rule._cb_reset_all(HABApp.openhab.events.ItemCommandEvent("name", test_case.event))

                self.assertEqual(len(test_case.manual_commands), shading_rule_mock._config.items.manual.oh_send_command.call_count)
                shading_rule_mock._config.items.manual.oh_send_command.assert_has_calls(test_case.manual_commands)


class TestSlatValueSun(tests.helper.test_case_base.TestCaseBase):
    """Tests for SlatValueSun."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Elevation", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_SlatValueSingle", None)  # NumberItem
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_SlatValueDual", None)  # DimmerItem

        self.config_single = habapp_rules.actors.config.shading.SlatValueConfig(
            items=habapp_rules.actors.config.shading.SlatValueItems(
                sun_elevation="Unittest_Elevation",
                slat_value="Unittest_SlatValueSingle",
            ),
            parameter=habapp_rules.actors.config.shading.SlatValueParameter(
                elevation_slat_characteristic=[
                    habapp_rules.actors.config.shading.ElevationSlatMapping(0, 100),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(10, 70),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(20, 60),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(30, 50),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(40, 50),
                ]
            ),
        )

        self.config_dual = habapp_rules.actors.config.shading.SlatValueConfig(
            items=habapp_rules.actors.config.shading.SlatValueItems(
                sun_elevation="Unittest_Elevation",
                summer="Unittest_Summer",
                slat_value="Unittest_SlatValueDual",
            ),
            parameter=habapp_rules.actors.config.shading.SlatValueParameter(
                elevation_slat_characteristic=copy.deepcopy(self.config_single.parameter.elevation_slat_characteristic),
                elevation_slat_characteristic_summer=[
                    habapp_rules.actors.config.shading.ElevationSlatMapping(0, 100),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(10, 80),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(20, 70),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(30, 60),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(40, 50),
                    habapp_rules.actors.config.shading.ElevationSlatMapping(50, 50),
                ],
            ),
        )

        self.slat_value_sun_single = habapp_rules.actors.shading.SlatValueSun(self.config_single)
        self.slat_value_sun_dual = habapp_rules.actors.shading.SlatValueSun(self.config_dual)

    def test__get_slat_value(self) -> None:
        """Test __get_slat_value."""
        tests.helper.oh_item.item_state_change_event("Unittest_Summer", "ON")
        TestCase = collections.namedtuple("TestCase", "elevation, expected_result_single, , expected_result_dual")

        test_cases = [
            TestCase(-10, 100, 100),
            TestCase(-1, 100, 100),
            TestCase(0, 100, 100),
            TestCase(1, 100, 100),
            TestCase(9, 100, 100),
            TestCase(10, 70, 80),
            TestCase(11, 70, 80),
            TestCase(29, 60, 70),
            TestCase(30, 50, 60),
            TestCase(31, 50, 60),
            TestCase(39, 50, 60),
            TestCase(40, 50, 50),
            TestCase(41, 50, 50),
            TestCase(60, 50, 50),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.assertEqual(test_case.expected_result_single, self.slat_value_sun_single._SlatValueSun__get_slat_value(test_case.elevation))
                self.assertEqual(test_case.expected_result_dual, self.slat_value_sun_dual._SlatValueSun__get_slat_value(test_case.elevation))

    def test_cb_elevation(self) -> None:
        """Test _cb_elevation."""
        with unittest.mock.patch.object(self.slat_value_sun_single, "_SlatValueSun__get_slat_value", side_effect=[42, 43]):
            tests.helper.oh_item.assert_value("Unittest_SlatValueSingle", None)
            tests.helper.oh_item.item_state_change_event("Unittest_Elevation", 10)
            tests.helper.oh_item.assert_value("Unittest_SlatValueSingle", 42)

            tests.helper.oh_item.item_state_change_event("Unittest_Elevation", 11)
            tests.helper.oh_item.assert_value("Unittest_SlatValueSingle", 43)

    def test_cb_summer_winter(self) -> None:
        """Test _cb_summer_winter."""
        # initial -> None == Winter
        tests.helper.oh_item.item_state_change_event("Unittest_Elevation", 30)
        self.assertEqual(self.config_dual.parameter.elevation_slat_characteristic, self.slat_value_sun_dual._slat_characteristic_active)
        tests.helper.oh_item.assert_value("Unittest_SlatValueDual", 50)

        # change to summer
        tests.helper.oh_item.item_state_change_event("Unittest_Summer", "ON")
        self.assertEqual(self.config_dual.parameter.elevation_slat_characteristic_summer, self.slat_value_sun_dual._slat_characteristic_active)
        tests.helper.oh_item.assert_value("Unittest_SlatValueDual", 60)

        # change to winter
        tests.helper.oh_item.item_state_change_event("Unittest_Summer", "OFF")
        self.assertEqual(self.config_dual.parameter.elevation_slat_characteristic, self.slat_value_sun_dual._slat_characteristic_active)
        tests.helper.oh_item.assert_value("Unittest_SlatValueDual", 50)


class TestReferenceRun(tests.helper.test_case_base.TestCaseBase):
    """Test ReferenceRun rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_TriggerRun", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_PresenceState", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "Unittest_LastRun", None)

        self.config = habapp_rules.actors.config.shading.ReferenceRunConfig(
            items=habapp_rules.actors.config.shading.ReferenceRunItems(
                trigger_run="Unittest_TriggerRun",
                presence_state="Unittest_PresenceState",
                last_run="Unittest_LastRun",
            )
        )

        self.rule = habapp_rules.actors.shading.ReferenceRun(self.config)

    def test_cb_presence_state(self) -> None:
        """Test _cb_presence_state."""
        TestCase = collections.namedtuple("TestCase", "now, last_run, presence_state, run_triggered")

        test_cases = [
            # trigger expected
            TestCase(datetime.datetime(2021, 1, 1, 10), datetime.datetime(2020, 12, 1, 10), PresenceState.ABSENCE, True),
            TestCase(datetime.datetime(2021, 2, 1, 10), datetime.datetime(2021, 1, 1, 10), PresenceState.ABSENCE, True),
            # no trigger -> wrong state
            TestCase(datetime.datetime(2021, 2, 1, 10), datetime.datetime(2021, 1, 1, 10), PresenceState.PRESENCE, False),
            TestCase(datetime.datetime(2021, 2, 1, 10), datetime.datetime(2021, 1, 1, 10), PresenceState.LEAVING, False),
            TestCase(datetime.datetime(2021, 2, 1, 10), datetime.datetime(2021, 1, 1, 10), PresenceState.LONG_ABSENCE, False),
            # no trigger -> already triggered this month
            TestCase(datetime.datetime(2021, 2, 1, 10), datetime.datetime(2021, 2, 1, 9), PresenceState.ABSENCE, False),
        ]

        with unittest.mock.patch("datetime.datetime") as datetime_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    tests.helper.oh_item.set_state("Unittest_LastRun", test_case.last_run)
                    tests.helper.oh_item.set_state("Unittest_TriggerRun", None)
                    datetime_mock.now.return_value = test_case.now

                    tests.helper.oh_item.item_state_change_event("Unittest_PresenceState", test_case.presence_state.value)

                    tests.helper.oh_item.assert_value("Unittest_TriggerRun", "ON" if test_case.run_triggered else None)



"""Test astro rules."""

import collections
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.sensors.astro
import habapp_rules.sensors.config.astro
import tests.helper.oh_item
import tests.helper.test_case_base


class TestSetDay(tests.helper.test_case_base.TestCaseBase):
    """Tests for TestSetDay."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Elevation", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Day", None)

    def test_init(self) -> None:
        """Test init without elevation."""
        # default threshold
        config = habapp_rules.sensors.config.astro.SetDayConfig(
            items=habapp_rules.sensors.config.astro.SetDayItems(
                day="Unittest_Day",
                elevation="Unittest_Elevation",
            )
        )

        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.soon") as run_soon_mock:
            rule = habapp_rules.sensors.astro.SetDay(config)

        run_soon_mock.assert_called_once_with(rule._set_night)
        self.assertEqual(0, rule._elevation_threshold)

        # custom threshold
        config.parameter = habapp_rules.sensors.config.astro.SetDayParameter(elevation_threshold=-2)
        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.soon") as run_soon_mock:
            rule = habapp_rules.sensors.astro.SetDay(config)

        run_soon_mock.assert_called_once_with(rule._set_night)
        self.assertEqual(-2, rule._elevation_threshold)

    def test_init_with_elevation(self) -> None:
        """Test init without elevation."""
        TestCase = collections.namedtuple("TestCase", "elevation_value, night_state")

        test_cases = [
            TestCase(None, None),
            TestCase(-1, "OFF"),
            TestCase(0, "OFF"),
            TestCase(0.9, "OFF"),
            TestCase(1, "OFF"),
            TestCase(1.1, "ON"),
            TestCase(2, "ON"),
            TestCase(10, "ON"),
        ]

        config = habapp_rules.sensors.config.astro.SetDayConfig(
            items=habapp_rules.sensors.config.astro.SetDayItems(
                day="Unittest_Day",
                elevation="Unittest_Elevation",
            ),
            parameter=habapp_rules.sensors.config.astro.SetDayParameter(elevation_threshold=1),
        )

        habapp_rules.sensors.astro.SetDay(config)

        for test_case in test_cases:
            tests.helper.oh_item.item_state_change_event("Unittest_Elevation", test_case.elevation_value)
            tests.helper.oh_item.assert_value("Unittest_Day", test_case.night_state)


class TestSetNight(tests.helper.test_case_base.TestCaseBase):
    """Tests for TestSetNight."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Elevation", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Night", None)

    def test_init(self) -> None:
        """Test init without elevation."""
        # default threshold
        config = habapp_rules.sensors.config.astro.SetNightConfig(
            items=habapp_rules.sensors.config.astro.SetNightItems(
                night="Unittest_Night",
                elevation="Unittest_Elevation",
            )
        )

        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.soon") as run_soon_mock:
            rule = habapp_rules.sensors.astro.SetNight(config)

        run_soon_mock.assert_called_once_with(rule._set_night)
        self.assertEqual(-8, rule._elevation_threshold)

        # custom threshold
        config.parameter = habapp_rules.sensors.config.astro.SetNightParameter(elevation_threshold=-10)
        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.soon") as run_soon_mock:
            rule = habapp_rules.sensors.astro.SetNight(config)

        run_soon_mock.assert_called_once_with(rule._set_night)
        self.assertEqual(-10, rule._elevation_threshold)

    def test_init_with_elevation(self) -> None:
        """Test init without elevation."""
        TestCase = collections.namedtuple("TestCase", "elevation_value, night_state")

        test_cases = [
            TestCase(None, None),
            TestCase(-9, "ON"),
            TestCase(-8.1, "ON"),
            TestCase(-8, "OFF"),
            TestCase(-7.9, "OFF"),
            TestCase(-5, "OFF"),
            TestCase(0, "OFF"),
            TestCase(10, "OFF"),
        ]

        config = habapp_rules.sensors.config.astro.SetNightConfig(
            items=habapp_rules.sensors.config.astro.SetNightItems(
                night="Unittest_Night",
                elevation="Unittest_Elevation",
            ),
            parameter=habapp_rules.sensors.config.astro.SetNightParameter(elevation_threshold=-8),
        )

        habapp_rules.sensors.astro.SetNight(config)

        for test_case in test_cases:
            tests.helper.oh_item.item_state_change_event("Unittest_Elevation", test_case.elevation_value)
            tests.helper.oh_item.assert_value("Unittest_Night", test_case.night_state)



"""Base classes for pydantic config models."""

import types
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.core.helper


class ItemBase(pydantic.BaseModel):
    """Base class for item config."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    @pydantic.model_validator(mode="before")
    @classmethod
    def check_all_fields_oh_items(cls, data: typing.Any) -> typing.Any:  # noqa: ANN401
        """Validate that all fields are OpenHAB items.

        All items must be subclasses of `HABApp.openhab.items.OpenhabItem`.
        If create_if_not_exists is set, only one type is allowed.
        For lists, only one type is allowed.

        Args:
            data: data object given by pydantic

        Returns:
            data object

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if validation fails
        """
        for name, field_info in cls.model_fields.items():
            field_types = cls._get_type_of_field(name)
            extra_args = extra if (extra := field_info.json_schema_extra) else {}

            if isinstance(field_types, types.GenericAlias):
                # type is list of OpenHAB items
                field_types = typing.get_args(field_types)[0]

                if isinstance(field_types, types.UnionType):
                    field_types = [arg for arg in typing.get_args(field_types) if arg is not types.NoneType]

                # validate that create_if_not_exists is not set for lists
                if extra_args.get("create_if_not_exists", False):
                    msg = "create_if_not_exists is not allowed for list fields"
                    raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

            field_types = field_types if isinstance(field_types, list) else [field_types]

            for field_type in field_types:
                if not issubclass(field_type, HABApp.openhab.items.OpenhabItem):
                    msg = f"Field {field_type} is not an OpenhabItem"
                    raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

            # validate that only one type is given if create_if_not_exists is set
            if extra_args.get("create_if_not_exists", False) and len(field_types) > 1:
                msg = "If create_if_not_exists is set, only one type is allowed"
                raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

        return data

    @pydantic.field_validator("*", mode="before")
    @classmethod
    def convert_to_oh_item(cls, var: str | HABApp.openhab.items.OpenhabItem, validation_info: pydantic.ValidationInfo) -> HABApp.openhab.items.OpenhabItem | None:
        """Convert to OpenHAB item.

        Args:
            var: variable given by the user
            validation_info: validation info given by pydantic

        Returns:
            variable converted to OpenHAB item

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if type is not supported
        """
        extra_args = extra if (extra := cls.model_fields[validation_info.field_name].json_schema_extra) else {}
        create_if_not_exists = extra_args.get("create_if_not_exists", False)

        if create_if_not_exists:
            item_type = cls._get_type_of_field(validation_info.field_name).__qualname__.removesuffix("Item")
            return habapp_rules.core.helper.create_additional_item(var, item_type)

        if isinstance(var, list):
            return [cls._get_oh_item(itm) for itm in var]

        if issubclass(type(var), HABApp.openhab.items.OpenhabItem) or isinstance(var, str):
            return cls._get_oh_item(var)

        if var is None:
            return None

        msg = f"The following var is not supported: {var}"
        raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

    @staticmethod
    def _get_oh_item(item: str | HABApp.openhab.items.OpenhabItem) -> HABApp.openhab.items.OpenhabItem:
        """Get OpenHAB item from string or OpenHAB item.

        Args:
            item: name of OpenHAB item or OpenHAB item

        Returns:
            OpenHAB item

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if type is not supported
        """
        return item if isinstance(item, HABApp.openhab.items.OpenhabItem) else HABApp.openhab.items.OpenhabItem.get_item(item)

    @classmethod
    def _get_type_of_field(cls, field_name: str) -> type | list[type]:
        """Get type of field.

        Args:
            field_name: name of field

        Returns:
            type of field, NoneType will be ignored
        """
        field_type = cls.model_fields[field_name].annotation
        if isinstance(field_type, types.UnionType):
            field_type = [arg for arg in typing.get_args(field_type) if arg is not types.NoneType]
        return field_type


class ParameterBase(pydantic.BaseModel):
    """Base class for parameter config."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)


class ConfigBase(pydantic.BaseModel):
    """Base class for config objects."""

    items: ItemBase | None
    parameter: ParameterBase | None

    def __init__(self, **data: typing.Any) -> None:  # noqa: ANN401
        """Initialize the model.

        Args:
            data: data object given by pydantic

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if validation fails
        """
        try:
            super().__init__(**data)
        except pydantic.ValidationError as exc:
            msg = f"Failed to validate model: {exc.errors()}"
            raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg) from exc



import collections
import unittest

import HABApp.openhab.items

import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.core.exceptions import HabAppRulesConfigurationError
from habapp_rules.energy.config.virtual_energy_meter import EnergyMeterBaseItems, EnergyMeterNumberConfig, EnergyMeterNumberItems, EnergyMeterNumberParameter, PowerMapping


class TestEnergyMeterBaseItems(unittest.TestCase):
    """Tests for EnergyMeterBaseItems."""

    def test_exceptions_with_missing_item(self) -> None:
        """Test exceptions with missing item."""
        TestCase = collections.namedtuple("TestCase", "power_item, energy_item, raises_exc")

        power_item = HABApp.openhab.items.NumberItem("Power")
        energy_item = HABApp.openhab.items.NumberItem("Energy")

        test_cases = [
            TestCase(power_item=None, energy_item=None, raises_exc=True),
            TestCase(power_item=None, energy_item=energy_item, raises_exc=False),
            TestCase(power_item=power_item, energy_item=None, raises_exc=False),
            TestCase(power_item=power_item, energy_item=energy_item, raises_exc=False),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                if test_case.raises_exc:
                    with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
                        EnergyMeterBaseItems(power_output=test_case.power_item, energy_output=test_case.energy_item)
                else:
                    EnergyMeterBaseItems(power_output=test_case.power_item, energy_output=test_case.energy_item)


class TestEnergyMeterNumberParameter(unittest.TestCase):
    """Tests for EnergyMeterNumberParameter."""

    def test_get_power(self) -> None:
        """Test get_power."""
        TestCase = collections.namedtuple("TestCase", "mapping, value, expected_result")

        mapping_1 = [PowerMapping(0, 0), PowerMapping(50, 500), PowerMapping(100, 1000)]
        mapping_2 = [PowerMapping(0, 5), PowerMapping(10, 20), PowerMapping(20, 40), PowerMapping(100, 1000)]
        mapping_3 = [PowerMapping(-10, 10), PowerMapping(10, -10)]

        test_cases = [
            # mapping 1
            TestCase(mapping=mapping_1, value=0, expected_result=0),
            TestCase(mapping=mapping_1, value=50, expected_result=500),
            TestCase(mapping=mapping_1, value=100, expected_result=1000),
            TestCase(mapping=mapping_1, value=75, expected_result=750),
            TestCase(mapping=mapping_1, value=-100, expected_result=0),
            TestCase(mapping=mapping_1, value=150, expected_result=1000),
            # mapping 2
            TestCase(mapping=mapping_2, value=0, expected_result=5),
            TestCase(mapping=mapping_2, value=5, expected_result=12.5),
            TestCase(mapping=mapping_2, value=20, expected_result=40),
            TestCase(mapping=mapping_2, value=50, expected_result=400),
            # mapping 3
            TestCase(mapping=mapping_3, value=-10, expected_result=10),
            TestCase(mapping=mapping_3, value=0, expected_result=0),
            TestCase(mapping=mapping_3, value=10, expected_result=-10),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                params = EnergyMeterNumberParameter(power_mapping=test_case.mapping)
                self.assertEqual(test_case.expected_result, params.get_power(test_case.value))

    def test_init_exceptions(self) -> None:
        """Test exceptions at initialization."""
        # mapping list too short
        with self.assertRaises(HabAppRulesConfigurationError):
            EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0)])


class TestEnergyMeterNumberConfig(tests.helper.test_case_base.TestCaseBase):
    """Test EnergyMeterNumberConfig."""

    def setUp(self) -> None:
        """Set up tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Power", None)

    def test_init_exceptions(self) -> None:
        """Test exceptions at initialization."""
        # value below min (with DimmerItem)
        with self.assertRaises(HabAppRulesConfigurationError):
            EnergyMeterNumberConfig(items=EnergyMeterNumberItems(monitored_item="Unittest_Dimmer", power_output="Unittest_Power"), parameter=EnergyMeterNumberParameter(power_mapping=[PowerMapping(-20, 0), PowerMapping(0, 100)]))

        # value above max (with DimmerItem)
        with self.assertRaises(HabAppRulesConfigurationError):
            EnergyMeterNumberConfig(items=EnergyMeterNumberItems(monitored_item="Unittest_Dimmer", power_output="Unittest_Power"), parameter=EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0), PowerMapping(101, 100)]))



import datetime
import unittest
import unittest.mock

import HABApp
import HABApp.openhab.definitions.helpers.persistence_data

import habapp_rules.energy.helper


class TestHelperFunctions(unittest.TestCase):
    """Test all global functions."""

    def test_get_historic_value(self) -> None:
        """Test _get_historic_value."""
        mock_item = unittest.mock.MagicMock()
        fake_persistence_data = HABApp.openhab.definitions.helpers.persistence_data.OpenhabPersistenceData()
        mock_item.get_persistence_data.return_value = fake_persistence_data

        start_time = datetime.datetime.now()
        end_time = start_time + datetime.timedelta(hours=1)

        # no data
        self.assertEqual(0, habapp_rules.energy.helper.get_historic_value(mock_item, start_time))
        mock_item.get_persistence_data.assert_called_once_with(start_time=start_time, end_time=end_time)

        # data
        fake_persistence_data.data = {"0.0": 42, "1.0": 1337}
        self.assertEqual(42, habapp_rules.energy.helper.get_historic_value(mock_item, start_time))



"""Base class for Rule with State Machine."""

import threading
import typing

import HABApp
import HABApp.openhab.connection.handler.func_sync
import transitions.extensions.states


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class StateMachineWithTimeout(transitions.Machine):
    """State machine class with timeout."""


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class HierarchicalStateMachineWithTimeout(transitions.extensions.HierarchicalMachine):
    """Hierarchical state machine class with timeout."""


class StateMachineRule(HABApp.Rule):
    """Base class for creating rules with a state machine."""

    states: typing.ClassVar[list[dict]] = []
    trans: typing.ClassVar[list[dict]] = []
    state: str

    def __init__(self, state_item: HABApp.openhab.items.StringItem) -> None:
        """Init rule with state machine.

        Args:
            state_item: name of the item to hold the state
        """
        self.state_machine: transitions.Machine | None = None
        HABApp.Rule.__init__(self)

        self._item_state = state_item

    def get_initial_log_message(self) -> str:
        """Get log message which can be logged at the init of a rule with a state machine.

        Returns:
            log message
        """
        return f"Init of rule '{self.__class__.__name__}' with name '{self.rule_name}' was successful. Initial state = '{self.state}' | State item = '{self._item_state.name}'"

    def _get_initial_state(self, default_value: str = "initial") -> str:
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._item_state.value and self._item_state.value in [item.get("name", None) for item in self.states if isinstance(item, dict)]:
            return self._item_state.value
        return default_value

    def _set_initial_state(self) -> None:
        """Set initial state.

        if the ``initial_state`` parameter of the state machine constructor is used the timeouts will not be started for the initial state.
        """
        self._set_state(self._get_initial_state())

    def _set_state(self, state_name: str) -> None:  # noqa: PLR6301
        """Set given state.

        Args:
            state_name: name of state
        """
        eval(f"self.to_{state_name}()")  # noqa: S307

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
        self._item_state.oh_send_command(self.state)

    def on_rule_removed(self) -> None:
        """Override this to implement logic that will be called when the rule has been unloaded."""
        # stop timeout timer of current state
        if self.state_machine:
            for itm in self.state_machine.get_state(self.state).runner.values():
                if isinstance(itm, threading.Timer) and itm.is_alive():
                    itm.cancel()



import HABApp
import pydantic

from habapp_rules.core.pydantic_base import ConfigBase, ItemBase, ParameterBase


class BathroomLightItems(ItemBase):
    """Items for bathroom light."""

    # lights
    light_main: HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="main light item")
    light_main_ctr: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="control item for main light, this can be used to detect switch on via dimming")
    light_main_color: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="main light color (Kelvin)")
    light_main_hcl: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="set HCL mode from KNX actor active for main light")
    light_mirror: HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="mirror light item")

    # environment
    sleeping_state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="sleeping state item")
    presence_state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="presence state item")

    # state machine
    manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to switch to manual mode and disable the automatic functions")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")


class BathroomLightParameter(ParameterBase):
    """Parameter for bathroom light."""

    color_mirror_sync: float = pydantic.Field(4000, description="color temperature for the mirror")
    min_brightness_mirror_sync: int = pydantic.Field(80, description="minimum brightness for main light if main and mirror light is ON")
    color_night: int = pydantic.Field(2600, description="color temperature for night mode")
    brightness_night: int = pydantic.Field(40, description="brightness for night mode")
    extended_sleep_time: int = pydantic.Field(15 * 60, description="additional sleep time in seconds", gt=0)
    brightness_night_extended: int | None = pydantic.Field(None, description="brightness for night mode extended")


class BathroomLightConfig(ConfigBase):
    """Config for bathroom light."""

    items: BathroomLightItems = pydantic.Field(..., description="items for the switch")
    parameter: BathroomLightParameter = pydantic.Field(BathroomLightParameter(), description="parameter for the switch")



"""Run all unit-tests."""

import logging
import pathlib
import sys
import unittest.mock

sys.path.append(str(pathlib.Path(__file__).resolve().parent.parent))

EXCLUDED_PY_FILES = ["run_unittest.py", "__init__.py", "rule_runner.py"]
INPUT_MODULES = [f"{'.'.join(f.parts)[:-3]}" for f in pathlib.Path("tests").rglob("*.py") if f.name not in EXCLUDED_PY_FILES]

logger_mock = unittest.mock.MagicMock()
logger_mock.level = logging.WARNING

with unittest.mock.patch("logging.getLogger", return_value=logger_mock):
    result = unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromNames(INPUT_MODULES))



"""Helper for testing transitions."""

import unittest.mock


def call_timeout(mock_object: unittest.mock.MagicMock) -> None:
    """Helper to simulate timeout of timer.

    Args:
        mock_object: mock object of transitions.extensions.states.Timer
    """
    timer_func = mock_object.call_args.args[1]
    timer_args = mock_object.call_args.kwargs.get("args", {})
    timer_func(*timer_args)



"""Config models for presence rules."""

import HABApp
import pydantic
from multi_notifier.connectors.connector_mail import Mail
from multi_notifier.connectors.connector_telegram import Telegram

import habapp_rules.core.pydantic_base


class NotificationItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for presence detection."""

    target_item: HABApp.openhab.items.OpenhabItem = pydantic.Field(..., description="Item which state change triggers a notification")


class NotificationParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for notification."""

    notify_connector: Mail | Telegram = pydantic.Field(..., description="Notifier which is used for sending notifications")
    recipients: str | list[str] = pydantic.Field(..., description="Recipients which should be notified")


class NotificationConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for notification."""

    items: NotificationItems = pydantic.Field(..., description="items for notification")
    parameter: NotificationParameter = pydantic.Field(..., description="parameter for notification")



import abc
import time

import HABApp

from habapp_rules.energy.config.virtual_energy_meter import EnergyMeterNumberConfig, EnergyMeterSwitchConfig


class _VirtualEnergyMeterBase(HABApp.Rule):
    """Base class for virtual energy meter classes."""

    def __init__(self, config: EnergyMeterSwitchConfig | EnergyMeterNumberConfig) -> None:
        HABApp.Rule.__init__(self)
        self._config = config
        self._monitored_item = config.items.monitored_switch if isinstance(config, EnergyMeterSwitchConfig) else config.items.monitored_item

        if self._config.items.energy_output is not None and self._config.items.energy_output.value is None:
            self._config.items.energy_output.oh_send_command(0)

        self._power = self._get_power()
        self._last_energy_countdown_reset = 0
        self._send_energy_countdown = self.run.countdown(self._get_energy_countdown_time(), self._cb_countdown_end)
        self._monitored_item.listen_event(self._cb_monitored_item, HABApp.openhab.events.ItemStateChangedEventFilter())

        if self._is_on():
            self.run.soon(self._cb_monitored_item, HABApp.openhab.events.ItemStateChangedEvent(self._monitored_item.name, self._monitored_item.value, None))

        if self._config.items.power_output is not None:
            self._config.items.power_output.oh_send_command(self._get_power() if self._is_on() else 0)

    @abc.abstractmethod
    def _get_power(self) -> float:
        """Get power for monitored, for ON state.

        Returns:
            power
        """

    @abc.abstractmethod
    def _is_on(self) -> bool:
        """Check if monitored item is on.

        Returns:
            True if monitored item is on
        """

    def _get_energy_countdown_time(self) -> float:
        """Get time to send energy.

        Returns:
            time to send energy in seconds
        """
        # calc time to send every X W (X from config)
        # E = P * t -> t = E / P -> t = E / P
        if self._power == 0:
            return 1  # avoid divide by zero

        return self._config.parameter.energy_update_resolution / self._power * 3_600_000

    def _cb_monitored_item(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa:  ARG002
        """Callback which is triggered if the monitored item changed.

        Args:
            event: event which triggered this callback
        """
        if self._is_on():
            self._power = self._get_power()

        if self._config.items.power_output is not None:
            self._config.items.power_output.oh_send_command(self._power if self._is_on() else 0)

    def _reset_countdown(self) -> None:
        """Reset countdown for sending energy."""
        self._send_energy_countdown.set_countdown(self._get_energy_countdown_time())
        self._send_energy_countdown.reset()
        self._last_energy_countdown_reset = time.time()

    def _cb_countdown_end(self) -> None:
        """Callback which is triggered if _send_energy_countdown ended."""
        self._update_energy_item(self._get_energy_countdown_time())
        self._reset_countdown()

    def _set_energy_from_remaining_time(self) -> None:
        """Callback which is triggered if switch / dimmer was switched off."""
        remaining_time = time.time() - self._last_energy_countdown_reset
        self._update_energy_item(remaining_time)
        self._send_energy_countdown.stop()

    def _update_energy_item(self, time_since_last_update: float) -> None:
        """Update energy item.

        Args:
            time_since_last_update: time since last update
        """
        new_energy_value = self._config.items.energy_output.value + self._power * time_since_last_update / 3_600_000
        self._config.items.energy_output.oh_send_command(new_energy_value)


class VirtualEnergyMeterSwitch(_VirtualEnergyMeterBase):
    """Rule to monitor energy consumption of switch items without a real energy meter.

    # Config
    config = VirtualEnergyMeterSwitch(config=EnergyMeterSwitchConfig(
        items=EnergyMeterSwitchItems(
            monitored_item="Switch",
            power_output="Virtual_Power",
            energy_output="Virtual_Energy"

        ),
        parameter=EnergyMeterSwitchParameter(
            power=42
        )
    ))

    # Rule init
    VirtualEnergyMeterSwitch(config)
    """

    def __init__(self, config: EnergyMeterSwitchConfig) -> None:
        """Init Rule.

        Args:
            config: Config for virtual energy meter
        """
        _VirtualEnergyMeterBase.__init__(self, config)

    def _get_power(self) -> float:
        """Get power for monitored, for ON state.

        Returns:
            power
        """
        return self._config.parameter.power

    def _is_on(self) -> bool:
        """Check if monitored item is on.

        Returns:
            True if monitored item is on
        """
        return self._monitored_item.is_on()

    def _cb_monitored_item(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the monitored item changed.

        Args:
            event: event which triggered this callback
        """
        super()._cb_monitored_item(event)

        if self._config.items.energy_output is not None:
            if event.value == "ON":
                self._reset_countdown()
            else:
                self._set_energy_from_remaining_time()


class VirtualEnergyMeterNumber(_VirtualEnergyMeterBase):
    """Rule to monitor energy consumption of dimmer / number items without a real energy meter.

    # Config
    config = VirtualEnergyMeterNumber(config=EnergyMeterNumberConfig(
        items=EnergyMeterNumberItems(
            monitored_item="Dimmer",
            power_output="Virtual_Power",
            energy_output="Virtual_Energy"

        ),
        parameter=EnergyMeterNumberParameter(
            power_mapping=[
                PowerMapping(value=0, power=0),
                PowerMapping(value=100, power=10_000)
            ]
        )
    ))

    # Rule init
    VirtualEnergyMeterNumber(config)
    """

    def __init__(self, config: EnergyMeterNumberConfig) -> None:
        """Init Rule.

        Args:
            config: Config for virtual energy meter
        """
        if config.items.monitored_item.value is None:
            config.items.monitored_item.oh_send_command(0)

        _VirtualEnergyMeterBase.__init__(self, config)

    def _get_power(self) -> float:
        """Get power for monitored, for ON state.

        Returns:
            power
        """
        return self._config.parameter.get_power(self._config.items.monitored_item.value)

    def _is_on(self) -> bool:
        """Check if monitored item is on.

        Returns:
            True if monitored item is on
        """
        return self._get_power() != 0

    def _cb_monitored_item(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the monitored item changed.

        Args:
            event: event which triggered this callback
        """
        super()._cb_monitored_item(event)

        if self._config.items.energy_output is not None:
            if not event.value or (event.value and event.old_value):
                # switch off or value changed
                self._set_energy_from_remaining_time()

            if event.value:
                # switch on or value changed
                self._reset_countdown()



"""Test Presence rule."""

import collections
import pathlib
import sys
import threading
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.presence
import habapp_rules.system.presence
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer


class TestPresence(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing presence rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door1", "CLOSED")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door2", "CLOSED")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Leaving", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Phone1", "ON")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Phone2", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "CustomState", "")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Presence_Unittest_Presence_state", "")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Presence", "ON")

        config = habapp_rules.system.config.presence.PresenceConfig(
            items=habapp_rules.system.config.presence.PresenceItems(presence="Unittest_Presence", leaving="Unittest_Leaving", outdoor_doors=["Unittest_Door1", "Unittest_Door2"], phones=["Unittest_Phone1", "Unittest_Phone2"], state="CustomState")
        )

        self._presence = habapp_rules.system.presence.Presence(config)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_Presence", None)
        tests.helper.oh_item.set_state("Unittest_Door1", None)
        tests.helper.oh_item.set_state("Unittest_Door2", None)
        tests.helper.oh_item.set_state("Unittest_Leaving", None)
        tests.helper.oh_item.set_state("Unittest_Phone1", None)
        tests.helper.oh_item.set_state("Unittest_Phone2", None)
        tests.helper.oh_item.set_state("CustomState", None)

        config = habapp_rules.system.config.presence.PresenceConfig(
            items=habapp_rules.system.config.presence.PresenceItems(presence="Unittest_Presence", leaving="Unittest_Leaving", outdoor_doors=["Unittest_Door1", "Unittest_Door2"], phones=["Unittest_Phone1", "Unittest_Phone2"], state="CustomState")
        )

        habapp_rules.system.presence.Presence(config)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        presence_graph = tests.helper.graph_machines.GraphMachineTimer(model=self._presence, states=self._presence.states, transitions=self._presence.trans, initial=self._presence.state, show_conditions=True)

        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Presence"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)
        presence_graph.get_graph().draw(picture_dir / "Presence.png", format="png", prog="dot")

    def test_minimal_init(self) -> None:
        """Test init with minimal set of arguments."""
        config = habapp_rules.system.config.presence.PresenceConfig(items=habapp_rules.system.config.presence.PresenceItems(presence="Unittest_Presence", leaving="Unittest_Leaving", state="CustomState"))

        presence_min = habapp_rules.system.presence.Presence(config)

        self.assertEqual([], presence_min._config.items.phones)
        self.assertEqual([], presence_min._config.items.outdoor_doors)

    def test_enums(self) -> None:
        """Test if all enums from __init__.py are implemented."""
        implemented_states = list(self._presence.state_machine.states)
        enum_states = [state.value for state in habapp_rules.system.PresenceState] + ["initial"]
        self.assertEqual(len(enum_states), len(implemented_states))
        self.assertTrue(all(state in enum_states for state in implemented_states))

    def test__init__(self) -> None:
        """Test init."""
        tests.helper.oh_item.assert_value("CustomState", "presence")
        self.assertEqual(self._presence.state, "presence")

    def test_get_initial_state(self) -> None:
        """Test getting correct initial state."""
        Testcase = collections.namedtuple("Testcase", "presence, outside_doors, leaving, phones, expected_result")

        testcases = [
            # presence ON | leaving OFF
            Testcase(presence="ON", leaving="OFF", outside_doors=[], phones=[], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=[], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=[], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="OFF", outside_doors=[], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["CLOSED"], phones=[], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["CLOSED"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN"], phones=[], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=[], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            # presence ON | leaving ON
            Testcase(presence="ON", leaving="ON", outside_doors=[], phones=[], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=[], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=[], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=[], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["CLOSED"], phones=[], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["CLOSED"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN"], phones=[], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN, CLOSED"], phones=[], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN, CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN, CLOSED"], phones=["OFF"], expected_result="leaving"),
            Testcase(presence="ON", leaving="ON", outside_doors=["OPEN, CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            # presence OFF | leaving OFF
            Testcase(presence="OFF", leaving="OFF", outside_doors=[], phones=[], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=[], phones=["ON"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=[], phones=["OFF"], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=[], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["CLOSED"], phones=[], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["CLOSED"], phones=["OFF"], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN"], phones=[], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN"], phones=["ON"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN"], phones=["OFF"], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN"], phones=["ON", "OFF"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=[], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["ON"], expected_result="presence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["OFF"], expected_result="absence"),
            Testcase(presence="OFF", leaving="OFF", outside_doors=["OPEN, CLOSED"], phones=["ON", "OFF"], expected_result="presence"),
            # all None
            Testcase(presence=None, leaving=None, outside_doors=[None, None], phones=[None, None], expected_result="default"),
        ]

        for testcase in testcases:
            with self.subTest(testcase=testcase):
                self._presence._config.items.presence.value = testcase.presence
                self._presence._config.items.leaving.value = testcase.leaving

                self._presence._config.items.outdoor_doors = [HABApp.openhab.items.ContactItem(f"Unittest_Door{idx}", state) for idx, state in enumerate(testcase.outside_doors)]
                self._presence._config.items.phones = [HABApp.openhab.items.SwitchItem(f"Unittest_Phone{idx}", state) for idx, state in enumerate(testcase.phones)]

                self.assertEqual(self._presence._get_initial_state("default"), testcase.expected_result, f"failed testcase: {testcase}")

    def test_get_initial_state_extra(self) -> None:
        """Test getting correct initial state for special cases."""
        # current state value is long_absence
        self._presence._config.items.presence.value = "OFF"
        self._presence._config.items.leaving.value = "OFF"
        self._presence._config.items.state.value = "long_absence"
        self._presence._config.items.outdoor_doors = []

        # no phones
        self._presence._config.items.phones = []
        self.assertEqual(self._presence._get_initial_state("default"), "long_absence")

        # with phones
        self._presence._config.items.phones = [HABApp.openhab.items.SwitchItem("Unittest_Phone1")]
        self.assertEqual(self._presence._get_initial_state("default"), "long_absence")

    def test_presence_trough_doors(self) -> None:
        """Test if outside doors set presence correctly."""
        tests.helper.oh_item.send_command("Unittest_Presence", "OFF")
        self._presence.state_machine.set_state("absence")
        self.assertEqual(self._presence.state, "absence")

        tests.helper.oh_item.send_command("Unittest_Door1", "CLOSED", "CLOSED")
        self.assertEqual(self._presence.state, "absence")

        tests.helper.oh_item.send_command("Unittest_Door1", "OPEN", "CLOSED")
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.assert_value("Unittest_Presence", "ON")

        tests.helper.oh_item.send_command("Unittest_Door1", "OPEN", "CLOSED")
        self.assertEqual(self._presence.state, "presence")

        tests.helper.oh_item.send_command("Unittest_Door1", "CLOSED", "CLOSED")
        self.assertEqual(self._presence.state, "presence")

    def test_normal_leaving(self) -> None:
        """Test if 'normal' leaving works correctly."""
        self._presence.state_machine.set_state("presence")
        self.assertEqual(self._presence.state, "presence")

        tests.helper.oh_item.send_command("Unittest_Leaving", "OFF", "ON")
        self.assertEqual(self._presence.state, "presence")

        tests.helper.oh_item.send_command("Unittest_Leaving", "ON", "OFF")
        self.assertEqual(self._presence.state, "leaving")
        self.transitions_timer_mock.assert_called_with(300, unittest.mock.ANY, args=unittest.mock.ANY)

        # call timeout and check if absence is active
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._presence.state, "absence")

        # leaving switches to on again -> state should be leaving again
        tests.helper.oh_item.send_command("Unittest_Leaving", "ON", "OFF")
        self.assertEqual(self._presence.state, "leaving")

        # test if also long absence is working
        self._presence.state = "long_absence"
        tests.helper.oh_item.send_command("Unittest_Leaving", "ON", "OFF")
        self.assertEqual(self._presence.state, "leaving")

    def test_abort_leaving(self) -> None:
        """Test aborting of leaving state."""
        self._presence.state_machine.set_state("presence")
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.set_state("Unittest_Leaving", "ON")

        tests.helper.oh_item.send_command("Unittest_Leaving", "ON", "OFF")
        self.assertEqual(self._presence.state, "leaving")
        tests.helper.oh_item.assert_value("Unittest_Leaving", "ON")

        tests.helper.oh_item.send_command("Unittest_Leaving", "OFF", "ON")
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.assert_value("Unittest_Leaving", "OFF")

    def test_abort_leaving_after_last_phone(self) -> None:
        """Test aborting of leaving which was started through last phone leaving."""
        self._presence.state_machine.set_state("presence")
        tests.helper.oh_item.set_state("Unittest_Phone1", "ON")

        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        tests.helper.timer.call_timeout(self.threading_timer_mock)
        self.assertEqual(self._presence.state, "leaving")
        tests.helper.oh_item.assert_value("Unittest_Leaving", "ON")

        tests.helper.oh_item.send_command("Unittest_Leaving", "OFF", "ON")
        self.assertEqual(self._presence.state, "presence")

        tests.helper.oh_item.send_command("Unittest_Phone1", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")

        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        tests.helper.timer.call_timeout(self.threading_timer_mock)
        self.assertEqual(self._presence.state, "leaving")
        tests.helper.oh_item.assert_value("Unittest_Leaving", "ON")

    def test_leaving_with_phones(self) -> None:
        """Test if leaving and absence is correct if phones appear/disappear during or after leaving."""
        # set initial states
        tests.helper.oh_item.set_state("Unittest_Phone1", "ON")
        tests.helper.oh_item.set_state("Unittest_Phone2", "OFF")
        self._presence.state_machine.set_state("presence")
        tests.helper.oh_item.send_command("Unittest_Leaving", "ON", "OFF")
        self.assertEqual(self._presence.state, "leaving")

        # leaving on, last phone disappears
        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        self.assertEqual(self._presence.state, "leaving")

        # leaving on, first phone appears
        tests.helper.oh_item.send_command("Unittest_Phone1", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")

        # leaving on, second phone appears
        tests.helper.oh_item.send_command("Unittest_Phone2", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")

        # leaving on, both phones leaving
        self._presence.state_machine.set_state("leaving")
        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        tests.helper.oh_item.send_command("Unittest_Phone2", "OFF", "ON")
        self.assertEqual(self._presence.state, "leaving")

        # absence on, one disappears, one stays online
        tests.helper.oh_item.send_command("Unittest_Phone1", "ON", "OFF")
        tests.helper.oh_item.send_command("Unittest_Phone2", "ON", "OFF")
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._presence.state, "absence")
        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        self.assertEqual(self._presence.state, "absence")

        # absence on, two phones disappears
        tests.helper.oh_item.send_command("Unittest_Phone2", "OFF", "ON")
        self.assertEqual(self._presence.state, "absence")

    def test__set_leaving_through_phone(self) -> None:
        """Test if leaving_detected is called correctly after timeout of __phone_absence_timer."""
        TestCase = collections.namedtuple("TestCase", "state, leaving_detected_called")

        test_cases = [TestCase("presence", True), TestCase("leaving", False), TestCase("absence", False), TestCase("long_absence", False)]

        for test_case in test_cases:
            with unittest.mock.patch.object(self._presence, "leaving_detected") as leaving_detected_mock:
                self._presence.state = test_case.state
                self._presence._Presence__set_leaving_through_phone()
            self.assertEqual(test_case.leaving_detected_called, leaving_detected_mock.called)

    def test_long_absence(self) -> None:
        """Test entering long_absence and leaving it."""
        # set initial state
        self._presence.state_machine.set_state("presence")
        tests.helper.oh_item.set_state("Unittest_Presence", "ON")

        # go to absence
        self._presence.absence_detected()
        self.assertEqual(self._presence.state, "absence")
        tests.helper.oh_item.assert_value("Unittest_Presence", "OFF")

        # check if timeout started, and stop the mocked timer
        self.transitions_timer_mock.assert_called_with(1.5 * 24 * 3600, unittest.mock.ANY, args=unittest.mock.ANY)
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._presence.state, "long_absence")
        tests.helper.oh_item.assert_value("Unittest_Presence", "OFF")

        # check if presence is set after door open
        self._presence._cb_outside_door(HABApp.openhab.events.ItemStateChangedEvent("Unittest_Door1", "OPEN", "CLOSED"))
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.assert_value("Unittest_Presence", "ON")

    def test_manual_change(self) -> None:
        """Test if change of presence object is setting correct state."""
        # send manual off from presence
        self._presence.state_machine.set_state("presence")
        tests.helper.oh_item.send_command("Unittest_Presence", "ON", "OFF")
        self._presence._cb_presence(HABApp.openhab.events.ItemStateChangedEvent("Unittest_Presence", "OFF", "ON"))
        self.assertEqual(self._presence.state, "absence")
        tests.helper.oh_item.send_command("Unittest_Presence", "OFF", "ON")

        # send manual off from leaving
        self._presence.state_machine.set_state("leaving")
        tests.helper.oh_item.send_command("Unittest_Presence", "ON", "OFF")
        self._presence._cb_presence(HABApp.openhab.events.ItemStateChangedEvent("Unittest_Presence", "OFF", "ON"))
        self.assertEqual(self._presence.state, "absence")
        tests.helper.oh_item.send_command("Unittest_Presence", "OFF", "ON")

        # send manual on from absence
        self._presence.state_machine.set_state("absence")
        tests.helper.oh_item.send_command("Unittest_Presence", "OFF", "ON")
        self._presence._cb_presence(HABApp.openhab.events.ItemStateChangedEvent("Unittest_Presence", "ON", "OFF"))
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.send_command("Unittest_Presence", "ON", "OFF")

        # send manual on from long_absence
        self._presence.state_machine.set_state("long_absence")
        tests.helper.oh_item.send_command("Unittest_Presence", "OFF", "ON")
        self._presence._cb_presence(HABApp.openhab.events.ItemStateChangedEvent("Unittest_Presence", "ON", "OFF"))
        self.assertEqual(self._presence.state, "presence")
        tests.helper.oh_item.send_command("Unittest_Presence", "ON", "OFF")

    def test_phones(self) -> None:
        """Test if presence is set correctly through phones."""
        # first phone switches to ON -> presence expected
        self._presence.state_machine.set_state("absence")
        tests.helper.oh_item.send_command("Unittest_Phone1", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")
        self.threading_timer_mock.assert_not_called()

        # second phone switches to ON -> no change expected
        tests.helper.oh_item.send_command("Unittest_Phone2", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")
        self.threading_timer_mock.assert_not_called()

        # second phone switches to OFF -> no change expected
        tests.helper.oh_item.send_command("Unittest_Phone2", "OFF", "ON")
        self.assertEqual(self._presence.state, "presence")
        self.threading_timer_mock.assert_not_called()

        # first phone switches to OFF -> timer should be started
        tests.helper.oh_item.send_command("Unittest_Phone1", "OFF", "ON")
        self.assertEqual(self._presence.state, "presence")
        self.threading_timer_mock.assert_called_once_with(1200, self._presence._Presence__set_leaving_through_phone)
        tests.helper.timer.call_timeout(self.threading_timer_mock)
        self.assertEqual(self._presence.state, "leaving")

        # phone appears during leaving -> leaving expected
        tests.helper.oh_item.send_command("Unittest_Phone1", "ON", "OFF")
        self.assertEqual(self._presence.state, "presence")
        self.assertIsNone(self._presence._Presence__phone_absence_timer)

        # timeout is over -> absence expected
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual(self._presence.state, "absence")

    def test_on_rule_removed(self) -> None:
        """Test on_rule_removed."""
        # timer NOT running
        with unittest.mock.patch("habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed") as parent_on_remove:
            self._presence.on_rule_removed()

        parent_on_remove.assert_called_once()

        # timer running
        self._presence._Presence__phone_absence_timer = threading.Timer(42, unittest.mock.MagicMock())
        self._presence._Presence__phone_absence_timer.start()
        with unittest.mock.patch("habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed") as parent_on_remove:
            self._presence.on_rule_removed()

        parent_on_remove.assert_called_once()
        self.assertIsNone(self._presence._Presence__phone_absence_timer)



"""Test KNX MQTT bridges."""

import collections
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.bridge.config.knx_mqtt
import habapp_rules.bridge.knx_mqtt
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base


class TestLight(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing Light rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_full_KNX_Dimmer_ctr", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_full_KNX_Switch_ctr", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_full_MQTT_dimmer", 0)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_switch_KNX_Switch_ctr", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_switch_MQTT_dimmer", 0)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_dimmer_KNX_Dimmer_ctr", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_dimmer_MQTT_dimmer", 0)

        config_full = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(
            items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_full_MQTT_dimmer", knx_switch_ctr="Unittest_full_KNX_Switch_ctr", knx_dimmer_ctr="Unittest_full_KNX_Dimmer_ctr")
        )

        config_switch = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_switch_MQTT_dimmer", knx_switch_ctr="Unittest_switch_KNX_Switch_ctr"))

        config_dimmer = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_dimmer_MQTT_dimmer", knx_dimmer_ctr="Unittest_dimmer_KNX_Dimmer_ctr"))

        self._knx_bridge_full = habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_full)
        self._knx_bridge_switch = habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_switch)
        self._knx_bridge_dimmer = habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_dimmer)

    def test__init__(self) -> None:
        """Test __init__."""
        self.assertIsNotNone(self._knx_bridge_full._config.items.knx_switch_ctr)
        self.assertIsNotNone(self._knx_bridge_full._config.items.knx_dimmer_ctr)

        self.assertIsNotNone(self._knx_bridge_switch._config.items.knx_switch_ctr)
        self.assertIsNone(self._knx_bridge_switch._config.items.knx_dimmer_ctr)

        self.assertIsNone(self._knx_bridge_dimmer._config.items.knx_switch_ctr)
        self.assertIsNotNone(self._knx_bridge_dimmer._config.items.knx_dimmer_ctr)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_full_MQTT_dimmer", None)
        tests.helper.oh_item.set_state("Unittest_full_KNX_Switch_ctr", None)
        tests.helper.oh_item.set_state("Unittest_full_KNX_Dimmer_ctr", None)
        tests.helper.oh_item.set_state("Unittest_switch_MQTT_dimmer", None)
        tests.helper.oh_item.set_state("Unittest_switch_KNX_Switch_ctr", None)
        tests.helper.oh_item.set_state("Unittest_dimmer_MQTT_dimmer", None)
        tests.helper.oh_item.set_state("Unittest_dimmer_KNX_Dimmer_ctr", None)

        config_full = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(
            items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_full_MQTT_dimmer", knx_switch_ctr="Unittest_full_KNX_Switch_ctr", knx_dimmer_ctr="Unittest_full_KNX_Dimmer_ctr")
        )

        config_switch = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_switch_MQTT_dimmer", knx_switch_ctr="Unittest_switch_KNX_Switch_ctr"))

        config_dimmer = habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_dimmer_MQTT_dimmer", knx_dimmer_ctr="Unittest_dimmer_KNX_Dimmer_ctr"))

        habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_full)
        habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_switch)
        habapp_rules.bridge.knx_mqtt.KnxMqttDimmerBridge(config_dimmer)

    def test_knx_on_off(self) -> None:
        """Test ON/OFF from KNX."""
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)

        # ON via KNX
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Switch_ctr", "ON")
        self.assertEqual(100, self._knx_bridge_full._config.items.mqtt_dimmer.value)

        # OFF via KNX
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Switch_ctr", "OFF")
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)

        # 50 via KNX
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", 50)
        self.assertEqual(50, self._knx_bridge_full._config.items.mqtt_dimmer.value)

        # 0 via KNX
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", 0)
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)

    def test_knx_increase(self) -> None:
        """Test increase from KNX."""
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", "INCREASE")
        self.assertEqual(60, self._knx_bridge_full._config.items.mqtt_dimmer.value)
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", "INCREASE")
        self.assertEqual(100, self._knx_bridge_full._config.items.mqtt_dimmer.value)

    def test_knx_decrease(self) -> None:
        """Test decrease from KNX."""
        self._knx_bridge_full._config.items.mqtt_dimmer.oh_send_command(100)
        self.assertEqual(100, self._knx_bridge_full._config.items.mqtt_dimmer.value)
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", "DECREASE")
        self.assertEqual(30, self._knx_bridge_full._config.items.mqtt_dimmer.value)
        tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", "DECREASE")
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)

    def test_knx_not_supported(self) -> None:
        """Test not supported command coming from KNX."""
        with unittest.mock.patch.object(self._knx_bridge_full, "_instance_logger") as logger_mock:
            tests.helper.oh_item.item_command_event("Unittest_full_KNX_Dimmer_ctr", "NotSupported")
            logger_mock.error.assert_called_once_with("command 'NotSupported' ist not supported!")

    def test_mqtt_events(self) -> None:
        """Test if KNX item is updated correctly if MQTT item changed."""
        self.assertEqual(0, self._knx_bridge_full._config.items.mqtt_dimmer.value)
        TestCase = collections.namedtuple("TestCase", "send_value, expected_call_dimmer, expected_call_switch")

        test_cases = [TestCase(70, 70, "ON"), TestCase(100, 100, "ON"), TestCase(1, 1, "ON"), TestCase(0, 0, "OFF")]

        with (
            unittest.mock.patch.object(self._knx_bridge_full._config.items, "knx_dimmer_ctr") as full_knx_dimmer_item_mock,
            unittest.mock.patch.object(self._knx_bridge_full._config.items, "knx_switch_ctr") as full_knx_switch_item_mock,
            unittest.mock.patch.object(self._knx_bridge_switch._config.items, "knx_switch_ctr") as switch_knx_switch_item_mock,
            unittest.mock.patch.object(self._knx_bridge_dimmer._config.items, "knx_dimmer_ctr") as dimmer_knx_dimmer_item_mock,
        ):
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    full_knx_dimmer_item_mock.oh_post_update.reset_mock()
                    full_knx_switch_item_mock.oh_post_update.reset_mock()
                    switch_knx_switch_item_mock.oh_post_update.reset_mock()
                    dimmer_knx_dimmer_item_mock.oh_post_update.reset_mock()

                    tests.helper.oh_item.item_state_change_event("Unittest_full_MQTT_dimmer", test_case.send_value)
                    tests.helper.oh_item.item_state_change_event("Unittest_switch_MQTT_dimmer", test_case.send_value)
                    tests.helper.oh_item.item_state_change_event("Unittest_dimmer_MQTT_dimmer", test_case.send_value)

                    # full bridge (switch and dimmer item for KNX)
                    full_knx_dimmer_item_mock.oh_post_update.assert_called_once_with(test_case.expected_call_dimmer)
                    full_knx_switch_item_mock.oh_post_update.assert_called_once_with(test_case.expected_call_switch)

                    # partial bridges
                    switch_knx_switch_item_mock.oh_post_update.assert_called_once_with(test_case.expected_call_switch)
                    dimmer_knx_dimmer_item_mock.oh_post_update.assert_called_once_with(test_case.expected_call_dimmer)



"""Unit tests for habapp_rules logger."""

import logging
import unittest
import unittest.mock

import habapp_rules.core.logger


class TestLoggerFunctions(unittest.TestCase):
    """Tests for all logger functions."""

    def test_setup_logger(self) -> None:
        """Test setup_logger."""
        stream_handler_mock = unittest.mock.MagicMock()
        file_handler_mock = unittest.mock.MagicMock()

        with unittest.mock.patch("logging.StreamHandler", return_value=stream_handler_mock), unittest.mock.patch("HABApp.config.logging.MidnightRotatingFileHandler", return_value=file_handler_mock):
            habapp_rules.core.logger.setup_logger()

            stream_handler_mock.setFormatter.assert_called_once()
            stream_handler_mock.setLevel.assert_called_once_with(logging.DEBUG)

            file_handler_mock.setFormatter.assert_called_once()
            file_handler_mock.setLevel.assert_called_once_with(logging.DEBUG)

            # path is existing
            with unittest.mock.patch("pathlib.Path.is_dir", return_value=False), unittest.mock.patch("pathlib.Path.mkdir") as makedirs_mock:
                habapp_rules.core.logger.setup_logger()
                makedirs_mock.assert_called_once_with(parents=True)

            # path is not existing
            with unittest.mock.patch("pathlib.Path.is_dir", return_value=True), unittest.mock.patch("pathlib.Path.mkdir") as makedirs_mock:
                habapp_rules.core.logger.setup_logger()
                makedirs_mock.assert_not_called()

        # remove handler
        habapp_rules_logger = logging.getLogger("habapp_rules")
        habapp_rules_logger.removeHandler(stream_handler_mock)
        habapp_rules_logger.removeHandler(file_handler_mock)


class TestInstanceLogger(unittest.TestCase):
    """Test for instanceLogger."""

    def test_instance_logger(self) -> None:
        """Test instance_logger."""
        instance_logger = habapp_rules.core.logger.InstanceLogger(logging.getLogger(__name__), "test_instance")

        self.assertEqual("test_instance", instance_logger._instance_name)
        self.assertEqual(("test_instance | test message", {}), instance_logger.process("test message", {}))



"""Test config models for KNX / MQTT bridge rules."""

import HABApp

import habapp_rules.bridge.config.knx_mqtt
import tests.helper.oh_item
import tests.helper.test_case_base


class TestKnxMqttConfig(tests.helper.test_case_base.TestCaseBase):
    """Test KnxMqttConfig."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_MQTT_dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_KNX_dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_KNX_switch", None)

    def test_validate_knx_items(self) -> None:
        """Test validate_knx_items."""
        # Both KNX items are given
        habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_MQTT_dimmer", knx_switch_ctr="Unittest_KNX_switch", knx_dimmer_ctr="Unittest_KNX_dimmer"))

        # only KNX switch is given
        habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(
            items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(
                mqtt_dimmer="Unittest_MQTT_dimmer",
                knx_switch_ctr="Unittest_KNX_switch",
            )
        )

        # only KNX dimmer is given
        habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(mqtt_dimmer="Unittest_MQTT_dimmer", knx_dimmer_ctr="Unittest_KNX_dimmer"))

        # no KNX item is given
        with self.assertRaises(ValueError):
            habapp_rules.bridge.config.knx_mqtt.KnxMqttConfig(
                items=habapp_rules.bridge.config.knx_mqtt.KnxMqttItems(
                    mqtt_dimmer="Unittest_MQTT_dimmer",
                )
            )
